#include <grub/machine/memory.h>
#include <grub/cpu/io.h>
#include <grub/lib/arg.h>
#include <grub/types.h>
#include <grub/dl.h>
#include <grub/serial.h>
#include <grub/i18n.h>
#include <grub/extcmd.h>
#include <grub/mm.h>
#include <grub/misc.h>
#include <grub/time.h>
#include <grub/buffalo_micon.h>

static struct buffalo_mes_set
{
	int num;
	char str[LCD_BUFFER_COL][LCD_BUFFER_ROW + 1];
} const err_sets[] =
{
	{4, {"SYSTEM Error E04", "Can't Load krnl!"}},
};


/* Argument options.  */
static const struct grub_arg_option micon_options[] =
{
  {"error_set", 'e', 0, N_("set error state"), 0, ARG_TYPE_INT},
  {0, 0, 0, 0, 0, 0}
};

static const struct grub_arg_option led_ctrl_options[] =
{
  {"power",   'p', 0, N_("Power LED"),            0, ARG_TYPE_NONE},
  {"info",    'i', 0, N_("Information LED"),      0, ARG_TYPE_NONE},
  {"ararm",   'a', 0, N_("Error LED"),            0, ARG_TYPE_NONE},
  {"link1",   '1', 0, N_("Link1 LED"),            0, ARG_TYPE_NONE},
  {"link2",   '2', 0, N_("Link2 LED"),            0, ARG_TYPE_NONE},
  {"execute", 'e', 0, N_("Execute LED"),          0, ARG_TYPE_NONE},
  {"disp",    'd', 0, N_("Disp LED"),             0, ARG_TYPE_NONE},
  {"red",     'r', 0, N_("LCD back light red"),   0, ARG_TYPE_NONE},
  {"green",   'g', 0, N_("LCD back light green"), 0, ARG_TYPE_NONE},
  {"blue",    'b', 0, N_("LCD back light blue"),  0, ARG_TYPE_NONE},
  {0, 0, 0, 0, 0, 0}
};

static const struct grub_arg_option sataled_ctrl_options[] =
{
  {"sata1e",  'a', 0, N_("SATA1 ERROR LED"),  0, ARG_TYPE_NONE},
  {"sata2e",  'b', 0, N_("SATA2 ERROR LED"),  0, ARG_TYPE_NONE},
  {"sata3e",  'c', 0, N_("SATA3 ERROR LED"),  0, ARG_TYPE_NONE},
  {"sata4e",  'd', 0, N_("SATA4 ERROR LED"),  0, ARG_TYPE_NONE},
  {"sata5e",  'e', 0, N_("SATA5 ERROR LED"),  0, ARG_TYPE_NONE},
  {"sata6e",  'f', 0, N_("SATA6 ERROR LED"),  0, ARG_TYPE_NONE},
  {"sata7e",  'g', 0, N_("SATA7 ERROR LED"),  0, ARG_TYPE_NONE},
  {"sata7e",  'h', 0, N_("SATA8 ERROR LED"),  0, ARG_TYPE_NONE},
  {"sata1",   '1', 0, N_("SATA1 LED"),        0, ARG_TYPE_NONE},
  {"sata2",   '2', 0, N_("SATA2 LED"),        0, ARG_TYPE_NONE},
  {"sata3",   '3', 0, N_("SATA3 LED"),        0, ARG_TYPE_NONE},
  {"sata4",   '4', 0, N_("SATA4 LED"),        0, ARG_TYPE_NONE},
  {"sata5",   '5', 0, N_("SATA5 LED"),        0, ARG_TYPE_NONE},
  {"sata6",   '6', 0, N_("SATA6 LED"),        0, ARG_TYPE_NONE},
  {"sata7",   '7', 0, N_("SATA7 LED"),        0, ARG_TYPE_NONE},
  {"sata7",   '8', 0, N_("SATA8 LED"),        0, ARG_TYPE_NONE},
  {0, 0, 0, 0, 0, 0}
};

static const struct grub_arg_option lcd_set_buff_options[] =
{
  {"first",  'f', 0, N_("Strings for first column."),  0, ARG_TYPE_STRING},
  {"second", 's', 0, N_("Strings for second column."), 0, ARG_TYPE_STRING},
  {0, 0, 0, 0, 0, 0}
};

static const struct grub_arg_option lcd_disp_ctrl_options[] =
{
  {"link",     'l', 0, N_("LCD item link speed"),   0, ARG_TYPE_NONE},
  {"capacity", 'c', 0, N_("LCD item Disk capacity"), 0, ARG_TYPE_NONE},
  {"host",     'h', 0, N_("LCD item hostname/IP address"), 0, ARG_TYPE_NONE},
  {"opmode",   'o', 0, N_("LCD item OP MODE"),  0, ARG_TYPE_NONE},
  {"date",     'd', 0, N_("LCD item Date"),     0, ARG_TYPE_NONE},
  {"buffer0",  '0', 0, N_("LCD item buffer 0"), 0, ARG_TYPE_NONE},
  {"buffer1",  '1', 0, N_("LCD item buffer 1"), 0, ARG_TYPE_NONE},
  {"buffer2",  '2', 0, N_("LCD item buffer 2"), 0, ARG_TYPE_NONE},
  {"buffer3",  '3', 0, N_("LCD item buffer 3"), 0, ARG_TYPE_NONE},
  {"buffer4",  '4', 0, N_("LCD item buffer 4"), 0, ARG_TYPE_NONE},
  {"buffer5",  '5', 0, N_("LCD item buffer 5"), 0, ARG_TYPE_NONE},
  {"buffer6",  '6', 0, N_("LCD item buffer 6"), 0, ARG_TYPE_NONE},
  {"buffer7",  '7', 0, N_("LCD item buffer 7"), 0, ARG_TYPE_NONE},
  {0, 0, 0, 0, 0, 0}
};

static const struct grub_arg_option led_brightness_options[] =
{
  {"backlight", 'b', 0, N_("Brightness of LCD backlight[0-15]"),   0, ARG_TYPE_INT},
  {"led",       'l', 0, N_("Brightness of LED[0-15]"), 0, ARG_TYPE_INT},
  {0, 0, 0, 0, 0, 0}
};

/* Serial port settings.  */
struct serial_port
{
  grub_port_t port;
  unsigned short divisor;
  unsigned short word_len;
  unsigned int   parity;
  unsigned short stop_bits;
};

/* Port settings for micon.  */
static struct serial_port uart_settings;

#ifdef GRUB_MACHINE_PCBIOS
static const unsigned short *uart_io_addr = (const unsigned short *) GRUB_MEMORY_MACHINE_BIOS_DATA_AREA_ADDR;
#define GRUB_UART_PORT_NUM 4
#else
#include <grub/i386/coreboot/serial.h>
static const grub_port_t uart_io_addr[] = GRUB_MACHINE_SERIAL_PORTS;
#define GRUB_UART_PORT_NUM (ARRAY_SIZE(uart_io_addr))
#endif

/* ========================================================================== */
/* =========================== UART ========================================= */
/* ========================================================================== */

/* Return the port number for the UNITth serial device.  */
static inline grub_port_t
uart_get_port (const unsigned int unit)
{
  if (unit < GRUB_UART_PORT_NUM)
    return uart_io_addr[unit];
  else
    return 0;
}

/* Convert speed to divisor.  */
static unsigned short
uart_get_divisor (unsigned int speed)
{
  unsigned int i;

  /* The structure for speed vs. divisor.  */
  struct divisor
  {
    unsigned int speed;
    unsigned short div;
  };

  /* The table which lists common configurations.  */
  /* 1843200 / (speed * 16)  */
  static struct divisor divisor_tab[] =
    {
      { 2400,   0x0030 },
      { 4800,   0x0018 },
      { 9600,   0x000C },
      { 19200,  0x0006 },
      { 38400,  0x0003 },
      { 57600,  0x0002 },
      { 115200, 0x0001 }
    };

  /* Set the baud rate.  */
  for (i = 0; i < sizeof (divisor_tab) / sizeof (divisor_tab[0]); i++)
    if (divisor_tab[i].speed == speed)
  /* UART in Yeeloong runs twice the usual rate.  */
#ifdef GRUB_MACHINE_MIPS_YEELOONG
      return 2 * divisor_tab[i].div;
#else
      return divisor_tab[i].div;
#endif
  return 0;
}

static grub_err_t
uart_init (void)
{
	unsigned char ier;

	ier = grub_inb(uart_settings.port + UART_IER);
	grub_outb(ier | UART_IER_RX_READY | UART_IER_TX_EMPTY,
		  uart_settings.port + UART_IER);

	grub_outb(UART_DLAB, uart_settings.port + UART_LCR);

	grub_outb(uart_settings.divisor & 0xff, uart_settings.port + UART_DLL);
	grub_outb((uart_settings.divisor >> 8) & 0xff, uart_settings.port + UART_DLH);

	grub_outb(uart_settings.word_len | uart_settings.parity | uart_settings.stop_bits,
		  uart_settings.port + UART_LCR);

	grub_outb(UART_ENABLE_FIFO, uart_settings.port + UART_FCR);

	return GRUB_ERR_NONE;
}

static void
uart_putc(const unsigned char c)
{
	while ((grub_inb(uart_settings.port + UART_LSR) & UART_EMPTY_TRANSMITTER) == 0);
	grub_outb(c, uart_settings.port + UART_TX);
}

static int
uart_getc(void)
{
	unsigned char c;
	unsigned int timeout = 10;

	do {
		grub_millisleep(1);
		if(--timeout == 0){
			return -1;
		}
	} while((grub_inb(uart_settings.port + UART_IIR) & UART_IIR_RX_READY) == 0);

	c = grub_inb(uart_settings.port + UART_RX);
	grub_dprintf("MICON", "recv 0x%02x\n", c); // for DEBUG

	return c;
}

static void
uart_puts(const unsigned char* buff, int len)
{
	int i = 0;

	while (len--) {
		uart_putc(buff[i++]);
	}
}

static grub_size_t
uart_gets(unsigned char* buff, int len)
{
	int i = 0, c;

	while (len--) {
		c = uart_getc();
		if (c < 0) {
			return -1;
		}
		buff[i++] = (unsigned char)c;
	}

	return (grub_size_t)i;
}

void
send_preamble(void)
{
	unsigned char nop[35];

	memset(nop, (unsigned char)0xff, sizeof(nop));
	uart_puts(nop, sizeof(nop));
}

int send_cmd (unsigned char op, unsigned char size, unsigned char command, unsigned char* data)
{
	int result, retry_cnt = MICON_RETRY_NUM;
	unsigned char tx_payload[(op == OP_WRITE ? size : 0) + 3];
	unsigned char rx_payload[(op == OP_READ ? size : 1) + 3];
	unsigned char checksum = 0x00;
	grub_size_t i;

	tx_payload[0] = op | (op == OP_WRITE ? size : 0x00);
	checksum += tx_payload[0];
	tx_payload[1] = command;
	checksum += tx_payload[1];
	if (op == OP_WRITE) {
		for (i = 0; i < size; i++) {
			tx_payload[2 + i] = data[i];
			checksum += tx_payload[2 + i];
		}
	}
	tx_payload[sizeof(tx_payload)-1] = 0x00 - checksum;

	do {
		result = 0;
		uart_puts(tx_payload, sizeof(tx_payload));
		/* for DEBUG
		for (i = 0; i < sizeof(tx_payload); i++) {
			grub_printf("send[%02d] = 0x%02x\n", i, tx_payload[i]);
		}
		*/

		if (uart_gets(rx_payload, sizeof(rx_payload)) != sizeof(rx_payload)) {
			result = -1;
			grub_printf("invalid micon rx payloads num\n");
			grub_error(GRUB_ERR_READ_ERROR, "invalid micon rx payloads num");
			goto OUT;
		}
		/* for DEBUG
		grub_printf("byte1 = 0x%02x\n", rx_payload[0]);
		grub_printf("byte2 = 0x%02x\n", rx_payload[1]);
		for (i = 2; i < sizeof(rx_payload)-1; i++) {
			grub_printf("data = 0x%02x\n", rx_payload[i]);
		}
		grub_printf("checksum = 0x%02x\n\n", rx_payload[sizeof(rx_payload)-1]);
		*/

		// SUM == 0 ?
		checksum = 0x00;
		for (i = 0; i < sizeof(rx_payload); i++) {
			checksum += rx_payload[i];
		}
		if (checksum != 0x00) {
			result = -1;
			grub_printf("SUM of micon rx payloads is not 0\n");
			grub_error(GRUB_ERR_READ_ERROR, "SUM of micon rx payloads is not 0");
			goto OUT;
		}

		// ACK == 0 ?
		if (op == OP_WRITE && rx_payload[2] != 0x00) {
			result = -1;
			grub_printf("ACK of micon rx payload is not 0\n");
			grub_error(GRUB_ERR_READ_ERROR, "ACK of micon rx payload is not 0");
			goto OUT;
		}

		if (op == OP_READ) {
			for (i = 0; i < size; i++) {
				data[i] = rx_payload[2 + i];
			}
		}
OUT:
		if (result == -1) {
			send_preamble();
		}
	} while(result != 0 && --retry_cnt > 0);

	return result;
}

/* ========================================================================== */
/* =========================== MICON ======================================== */
/* ========================================================================== */

grub_err_t
micon_bz_on (unsigned char sound)
{
	unsigned char data[DT_SIZE_1];

	data[0] = sound;

	return send_cmd(OP_WRITE, DT_SIZE_1, OP_CODE_BZ_ON, data);
}

grub_err_t
micon_led_ctrl (unsigned char state, unsigned short bits)
{
	grub_size_t i;
	unsigned char mask[DT_SIZE_2];
	unsigned char data[DT_SIZE_2];

	for (i = 0; i < sizeof(mask); i++) {
		mask[i] = (bits >> (8 * i)) & 0x00ff;
	}

	// change LED controller
	if (send_cmd(OP_READ, DT_SIZE_2, OP_CODE_LED_CPU_MCON, data) != 0) {
		return -1;
	}
	for (i = 0; i < sizeof(data); i++) {
		data[i] |= mask[i];
	}
	if (send_cmd(OP_WRITE, DT_SIZE_2, OP_CODE_LED_CPU_MCON, data) != 0) {
		return -1;
	}

	// change blink ON/OFF
	if (send_cmd(OP_READ, DT_SIZE_2, OP_CODE_LED_BLINK, data) != 0) {
		return -1;
	}
	for (i = 0; i < sizeof(data); i++) {
		if (state == STATE_BLINK) {
			data[i] |= mask[i];
		} else {
			data[i] &= ~mask[i];
		}
	}
	if (send_cmd(OP_WRITE, DT_SIZE_2, OP_CODE_LED_BLINK, data) != 0) {
		return -1;
	}

	// change ON/OFF
	if (send_cmd(OP_READ, DT_SIZE_2, OP_CODE_LED_ON_OFF, data) != 0) {
		return -1;
	}
	for (i = 0; i < sizeof(data); i++) {
		if (state == STATE_OFF) {
			data[i] &= ~mask[i];
		} else {
			data[i] |= mask[i];
		}
	}
	if (send_cmd(OP_WRITE, DT_SIZE_2, OP_CODE_LED_ON_OFF, data) != 0) {
		return -1;
	}

	/*
	// restore LED controller
	if (send_cmd(OP_READ, DT_SIZE_2, OP_CODE_LED_CPU_MCON, data) != 0) {
		return -1;
	}
	for (i = 0; i < sizeof(data); i++) {
		data[i] &= ~mask[i];
	}
	if (send_cmd(OP_WRITE, DT_SIZE_2, OP_CODE_LED_CPU_MCON, data) != 0) {
		return -1;
	}
	*/

	return GRUB_ERR_NONE;
}

grub_err_t
micon_sataled_ctrl (unsigned char state, unsigned short bits)
{
	grub_size_t i;
	unsigned char mask[DT_SIZE_2];
	unsigned char data[DT_SIZE_2];

	for (i = 0; i < sizeof(mask); i++) {
		mask[i] = (bits >> (8 * i)) & 0x00ff;
	}

	// change blink ON/OFF
	if (send_cmd(OP_READ, DT_SIZE_2, OP_CODE_SATA_LED_BLINK, data) != 0) {
		return -1;
	}
	for (i = 0; i < sizeof(data); i++) {
		if (state == STATE_BLINK) {
			data[i] |= mask[i];
		} else {
			data[i] &= ~mask[i];
		}
	}
	if (send_cmd(OP_WRITE, DT_SIZE_2, OP_CODE_SATA_LED_BLINK, data) != 0) {
		return -1;
	}

	// change ON/OFF
	if (send_cmd(OP_READ, DT_SIZE_2, OP_CODE_SATA_LED_ON_OFF, data) != 0) {
		return grub_error (GRUB_ERR_READ_ERROR, "micon read error");
	}
	for (i = 0; i < sizeof(data); i++) {
		if (state == STATE_OFF) {
			data[i] &= ~mask[i];
		} else {
			data[i] |= mask[i];
		}
	}
	if (send_cmd(OP_WRITE, DT_SIZE_2, OP_CODE_SATA_LED_ON_OFF, data) != 0) {
		return grub_error (GRUB_ERR_READ_ERROR, "micon read error");
	}

	return GRUB_ERR_NONE;
}

grub_err_t
micon_lcd_set_buff (unsigned int bufnum, unsigned char *data)
{
	return send_cmd(OP_WRITE, DT_SIZE_32, OP_CODE_LCD_DISP_BUF_BASE + bufnum, data);
}

grub_err_t
micon_lcd_disp_ctrl (unsigned char state, unsigned short bits)
{
	grub_size_t i;
	unsigned char mask[DT_SIZE_2];
	unsigned char data[DT_SIZE_2];

	for (i = 0; i < sizeof(mask); i++) {
		mask[i] = (bits >> (8 * i)) & 0x00ff;
	}

	if (send_cmd(OP_READ, DT_SIZE_1, OP_CODE_LCD_DISP_ON_OFF, &data[0]) != 0) {
		return -1;
	}
	if (send_cmd(OP_READ, DT_SIZE_1, OP_CODE_LCD_DISP_ON_OFF_EX, &data[1]) != 0) {
		return -1;
	}

	for (i = 0; i < sizeof(data); i++) {
		if (state == STATE_ON) {
			data[i] |= mask[i];
		} else {
			data[i] &= ~mask[i];
		}
	}

	if (send_cmd(OP_WRITE, DT_SIZE_1, OP_CODE_LCD_DISP_ON_OFF, &data[0]) != 0) {
		return -1;
	}
	if (send_cmd(OP_WRITE, DT_SIZE_1, OP_CODE_LCD_DISP_ON_OFF_EX, &data[1]) != 0) {
		return -1;
	}

	return GRUB_ERR_NONE;
}

grub_err_t
micon_lcd_disp_item (unsigned char item)
{
	return send_cmd(OP_WRITE, DT_SIZE_0, OP_CODE_DISP_BASE + item, NULL);
}

grub_err_t
micon_led_brightness (unsigned char lcd, unsigned char led)
{
	unsigned char data[DT_SIZE_1];

	if (led == 0xff && lcd == 0xff) {
		if (send_cmd(OP_READ, DT_SIZE_1, OP_CODE_LED_BRIGHTNESS, data) != 0) {
			return -1;
		}
		grub_printf("LCD Backlight brightness: %02d\n", ((data[0] >> 4) & 0x0f));
		grub_printf("LED brightness: %02d\n", data[0] & 0x0f);
		return GRUB_ERR_NONE;
	} else {
		data[0] = ((lcd << 4) & 0xf0) | (led & 0x0f);
		return send_cmd(OP_WRITE, DT_SIZE_1, OP_CODE_LED_BRIGHTNESS, data);
	}
}

/* ========================================================================== */
/* =========================== GRUB ========================================= */
/* ========================================================================== */

static grub_err_t
grub_cmd_bz_on (grub_extcmd_t cmd __attribute__((unused)), int argc, char** args)
{
	unsigned char data;

	if (argc > 1) {
		return grub_error (GRUB_ERR_BAD_ARGUMENT, "too many args");
	} else if (argc < 1) {
		return grub_error (GRUB_ERR_BAD_ARGUMENT, "too few args");
	}

	if (grub_strcmp(args[0], "stop") == 0) {
		data = DT_BZ_STOP;
	} else if (grub_strcmp(args[0], "machine") == 0) {
		data = DT_BZ_MACHINE;
	} else if (grub_strcmp(args[0], "button") == 0) {
		data = DT_BZ_BUTTON;
	} else if (grub_strcmp(args[0], "continue") == 0) {
		data = DT_BZ_CONTINUE;
	} else if (grub_strcmp(args[0], "switch1") == 0) {
		data = DT_BZ_SWITCH1;
	} else if (grub_strcmp(args[0], "switch2") == 0) {
		data = DT_BZ_SWITCH2;
	} else if (grub_strcmp(args[0], "music1") == 0) {
		data = DT_BZ_MUSIC1;
	} else if (grub_strcmp(args[0], "music2") == 0) {
		data = DT_BZ_MUSIC2;
	} else {
		return grub_error (GRUB_ERR_BAD_ARGUMENT, "unknown arg");
	}

	if (micon_bz_on(data) != 0) {
		return grub_error (GRUB_ERR_BAD_DEVICE, "micon I/O error");
	}
	return GRUB_ERR_NONE;
}

static grub_err_t
grub_cmd_led_ctrl (grub_extcmd_t cmd, int argc, char** args)
{
	grub_size_t i;
	unsigned char state;
	unsigned short mask = 0x0000;
	unsigned short bits[] = {
		DT_LED_POWER,
		DT_LED_INFO,
		DT_LED_ALARM,
		DT_LED_LINK1,
		DT_LED_LINK2,
		DT_LED_EXECUTE,
		DT_LED_DISP,
		DT_LED_RED,
		DT_LED_GREEN,
		DT_LED_BLUE,
	}; // Element of this array is related to led_ctrl_options

	if (argc > 1) {
		return grub_error (GRUB_ERR_BAD_ARGUMENT, "too many args");
	} else if (argc < 1) {
		return grub_error (GRUB_ERR_BAD_ARGUMENT, "too few args");
	}

	if (grub_strcmp(args[0], "on") == 0) {
		state = STATE_ON;
	} else if (grub_strcmp(args[0], "off") == 0) {
		state = STATE_OFF;
	} else if (grub_strcmp(args[0], "blink") == 0) {
		state = STATE_BLINK;
	} else {
		return grub_error (GRUB_ERR_BAD_ARGUMENT, "unknown arg");
	}

	for (i = 0; i < (sizeof(bits)/sizeof(unsigned short)); i++) {
		if (cmd->state[i].set) {
			mask |= bits[i];
		}
	}

	if (micon_led_ctrl(state, mask) != 0) {
		return grub_error (GRUB_ERR_BAD_DEVICE, "micon I/O error");
	}

	return GRUB_ERR_NONE;
}

static grub_err_t
grub_cmd_sataled_ctrl (grub_extcmd_t cmd, int argc, char** args)
{
	grub_size_t i;
	unsigned char state;
	unsigned short mask = 0x0000;
	unsigned short bits[] = {
		DT_SATA_LED_1_E,
		DT_SATA_LED_2_E,
		DT_SATA_LED_3_E,
		DT_SATA_LED_4_E,
		DT_SATA_LED_5_E,
		DT_SATA_LED_6_E,
		DT_SATA_LED_7_E,
		DT_SATA_LED_8_E,
		DT_SATA_LED_1,
		DT_SATA_LED_2,
		DT_SATA_LED_3,
		DT_SATA_LED_4,
		DT_SATA_LED_5,
		DT_SATA_LED_6,
		DT_SATA_LED_7,
		DT_SATA_LED_8,
	}; // Element of this array is related to sataled_ctrl_options

	if (argc > 1) {
		return grub_error (GRUB_ERR_BAD_ARGUMENT, "too many args");
	} else if (argc < 1) {
		return grub_error (GRUB_ERR_BAD_ARGUMENT, "too few args");
	}

	if (grub_strcmp(args[0], "on") == 0) {
		state = STATE_ON;
	} else if (grub_strcmp(args[0], "off") == 0) {
		state = STATE_OFF;
	} else if (grub_strcmp(args[0], "blink") == 0) {
		state = STATE_BLINK;
	} else {
		return grub_error (GRUB_ERR_BAD_ARGUMENT, "unknown arg");
	}

	for (i = 0; i < (sizeof(bits)/sizeof(unsigned short)); i++) {
		if (cmd->state[i].set) {
			mask |= bits[i];
		}
	}

	if (micon_sataled_ctrl(state, mask) != 0) {
		return grub_error (GRUB_ERR_BAD_DEVICE, "micon I/O error");
	}

	return GRUB_ERR_NONE;
}

static grub_err_t
grub_cmd_lcd_set_buff (grub_extcmd_t cmd, int argc, char** args)
{
	int i, column_len;
	unsigned int bufnum;
	unsigned char data[LCD_BUFFER_ROW * LCD_BUFFER_COL];

	if (argc > 1) {
		return grub_error (GRUB_ERR_BAD_ARGUMENT, "too many args");
	} else if (argc < 1) {
		return grub_error (GRUB_ERR_BAD_ARGUMENT, "too few args");
	}

	if (grub_strlen(args[0]) != 1 && (args[0][0] < '0' || '7' < args[0][0])) {
		return grub_error (GRUB_ERR_BAD_ARGUMENT, "invalid buffer number");
	}
	bufnum = args[0][0] - '0';

	if (!cmd->state[0].set && !cmd->state[1].set) {
		return grub_error (GRUB_ERR_BAD_ARGUMENT, "no strings");
	}

	memset(data, 0x20, sizeof(data));
	for (i = 0; i < LCD_BUFFER_COL; i++) {
		if (cmd->state[i].set) {
			column_len = grub_strlen(cmd->state[i].arg);
			if (column_len > LCD_BUFFER_ROW) {
				column_len = LCD_BUFFER_ROW;
			}
			memcpy(&data[LCD_BUFFER_ROW * i], cmd->state[i].arg, column_len);
		}
	}

	if (micon_lcd_set_buff(bufnum, data) != 0) {
		return grub_error (GRUB_ERR_BAD_DEVICE, "micon I/O error");
	}

	return GRUB_ERR_NONE;
}

static grub_err_t
grub_cmd_lcd_disp_ctrl (grub_extcmd_t cmd, int argc, char** args)
{
	grub_size_t i;
	unsigned char state;
	unsigned short mask = 0x0000;
	unsigned short bits[] = {
		DT_DISP_LINK,
		DT_DISP_DISK,
		DT_DISP_HOST,
		DT_DISP_OP,
		DT_DISP_DATE,
		DT_DISP_BUFF0,
		DT_DISP_BUFF1,
		DT_DISP_BUFF2,
		DT_DISP_BUFF3,
		DT_DISP_BUFF4,
		DT_DISP_BUFF5,
		DT_DISP_BUFF6,
		DT_DISP_BUFF7,
	}; // Element of this array is related to lcd_disp_ctrl_options

	if (argc > 1) {
		return grub_error (GRUB_ERR_BAD_ARGUMENT, "too many args");
	} else if (argc < 1) {
		return grub_error (GRUB_ERR_BAD_ARGUMENT, "too few args");
	}

	if (grub_strcmp(args[0], "on") == 0) {
		state = STATE_ON;
	} else if (grub_strcmp(args[0], "off") == 0) {
		state = STATE_OFF;
	} else {
		return grub_error (GRUB_ERR_BAD_ARGUMENT, "unknown arg");
	}

	for (i = 0; i < (sizeof(bits)/sizeof(unsigned short)); i++) {
		if (cmd->state[i].set) {
			mask |= bits[i];
		}
	}

	if (micon_lcd_disp_ctrl(state, mask) != 0) {
		return grub_error (GRUB_ERR_BAD_DEVICE, "micon I/O error");
	}

	return GRUB_ERR_NONE;
}

static grub_err_t
grub_cmd_lcd_disp_item (grub_extcmd_t cmd __attribute__((unused)), int argc, char** args)
{
	unsigned char op_code;
	
	if (argc > 1) {
		return grub_error (GRUB_ERR_BAD_ARGUMENT, "too many args");
	} else if (argc < 1) {
		return grub_error (GRUB_ERR_BAD_ARGUMENT, "too few args");
	}

	if (grub_strcmp(args[0], "link") == 0) {
		op_code = OP_CODE_DISP_LINK;
	} else if (grub_strcmp(args[0], "hostname") == 0) {
		op_code = OP_CODE_DISP_HOSTNAME;
	} else if (grub_strcmp(args[0], "ipaddress") == 0) {
		op_code = OP_CODE_DISP_IPADDRESS;
	} else if (grub_strcmp(args[0], "opmode") == 0) {
		op_code = OP_CODE_DISP_OPMODE;
	} else if (grub_strcmp(args[0], "date") == 0) {
		op_code = OP_CODE_DISP_DATE;
	} else if (grub_strcmp(args[0], "buffer0") == 0) {
		op_code = OP_CODE_DISP_BUFF0;
	} else if (grub_strcmp(args[0], "buffer1") == 0) {
		op_code = OP_CODE_DISP_BUFF1;
	} else if (grub_strcmp(args[0], "buffer2") == 0) {
		op_code = OP_CODE_DISP_BUFF2;
	} else if (grub_strcmp(args[0], "buffer3") == 0) {
		op_code = OP_CODE_DISP_BUFF3;
	} else if (grub_strcmp(args[0], "buffer4") == 0) {
		op_code = OP_CODE_DISP_BUFF4;
	} else if (grub_strcmp(args[0], "buffer5") == 0) {
		op_code = OP_CODE_DISP_BUFF5;
	} else if (grub_strcmp(args[0], "buffer6") == 0) {
		op_code = OP_CODE_DISP_BUFF6;
	} else if (grub_strcmp(args[0], "buffer7") == 0) {
		op_code = OP_CODE_DISP_BUFF7;
	} else {
		return grub_error (GRUB_ERR_BAD_ARGUMENT, "unknown arg");
	}

	if (micon_lcd_disp_item(op_code) != 0) {
		return grub_error (GRUB_ERR_BAD_DEVICE, "micon I/O error");
	}

	return GRUB_ERR_NONE;
}

static grub_err_t
grub_cmd_led_brightness (grub_extcmd_t cmd, int argc __attribute__ ((unused)), char** args __attribute__ ((unused)))
{
	unsigned char lcd, led;

	lcd = led = 0xff;
	if (cmd->state[0].set) {
		lcd = grub_strtoul(cmd->state[0].arg, 0, 0);
	}
	if (cmd->state[1].set) {
		led = grub_strtoul(cmd->state[1].arg, 0, 0);
	}

	if (micon_led_brightness(lcd, led) != 0) {
		return grub_error (GRUB_ERR_BAD_DEVICE, "micon I/O error");
	}
	return GRUB_ERR_NONE;
}

static grub_err_t
grub_cmd_micon (grub_extcmd_t cmd, int argc __attribute__ ((unused)), char** args __attribute__ ((unused)))
{
	unsigned int i, j;
	unsigned char data[DT_SIZE_32];

	if (cmd->state[0].set) {
		int err_num = grub_strtol(cmd->state[0].arg, 0, 0);
		unsigned short mask = 0x0000;

		// LCD
		memset(data, 0x20, sizeof(data));
		for (i = 0; i < sizeof(err_sets)/sizeof(struct buffalo_mes_set); i++) {
			if (err_sets[i].num == err_num) {
				for (j = 0; j < LCD_BUFFER_COL; j++) {
					memcpy(&data[LCD_BUFFER_ROW * j],
					       err_sets[i].str[j],
					       grub_strlen(err_sets[i].str[j]));
				}
				break;
			}
		}
		if (micon_lcd_set_buff(1, data) != 0) {
			return grub_error (GRUB_ERR_BAD_DEVICE, "micon I/O error");
		}

		// force display buffer1
		mask = DT_DISP_BUFF1;
		if (micon_lcd_disp_ctrl(STATE_ON, mask) != 0) {
			return grub_error (GRUB_ERR_BAD_DEVICE, "micon I/O error");
		}
		if (micon_lcd_disp_item(OP_CODE_DISP_BUFF1) != 0) {
			return grub_error (GRUB_ERR_BAD_DEVICE, "micon I/O error");
		}
		mask = DT_DISP_BUFF0;
		if (micon_lcd_disp_ctrl(STATE_OFF, mask) != 0) {
			return grub_error (GRUB_ERR_BAD_DEVICE, "micon I/O error");
		}

		// SATA LED
		mask = 0xffff; // ALL
		if (micon_sataled_ctrl(STATE_OFF, mask) != 0) {
			return grub_error (GRUB_ERR_BAD_DEVICE, "micon I/O error");
		}

		// LED
		mask = DT_LED_GREEN | DT_LED_BLUE;
		if (micon_led_ctrl(STATE_OFF, mask) != 0) {
			return grub_error (GRUB_ERR_BAD_DEVICE, "micon I/O error");
		}
		mask = DT_LED_RED | DT_LED_ALARM;
		if (micon_led_ctrl(STATE_ON, mask) != 0) {
			return grub_error (GRUB_ERR_BAD_DEVICE, "micon I/O error");
		}

		// LED brightness
		if (micon_led_brightness(DT_BRIGHTNESS_MAX, DT_BRIGHTNESS_MAX) != 0) {
			return grub_error (GRUB_ERR_BAD_DEVICE, "micon I/O error");
		}

		// buzzer
		if (micon_bz_on(DT_BZ_SWITCH2) != 0) {
			return grub_error (GRUB_ERR_BAD_DEVICE, "micon I/O error");
		}
	}

	return GRUB_ERR_NONE;
}


/* ========================================================================== */
/* =========================== GRUB ========================================= */
/* ========================================================================== */

static grub_extcmd_t cmd;
static grub_extcmd_t cmd_bz_on;
static grub_extcmd_t cmd_led_ctrl;
static grub_extcmd_t cmd_sataled_ctrl;
static grub_extcmd_t cmd_lcd_set_buff;
static grub_extcmd_t cmd_lcd_disp_ctrl;
static grub_extcmd_t cmd_lcd_disp_item;

GRUB_MOD_INIT(buffalo_micon)
{
	cmd = grub_register_extcmd("micon", grub_cmd_micon,
				   GRUB_COMMAND_FLAG_BOTH,
				   "micon [-e error_num]","micon access",
				   micon_options);
	
	cmd_bz_on =
		grub_register_extcmd("bz_on", grub_cmd_bz_on,
				     GRUB_COMMAND_FLAG_BOTH,
				     "bz_on stop|machine|button|continue|switch1|switch2|music1|music2",
				     "Control buzzer.", 0);
	cmd_led_ctrl =
		grub_register_extcmd("led_ctrl", grub_cmd_led_ctrl,
				     GRUB_COMMAND_FLAG_BOTH,
				     "led_ctrl [-pia12edrgb] on|off|blink",
				     "Turn on/off LED", led_ctrl_options);

	cmd_sataled_ctrl =
		grub_register_extcmd("sataled_ctrl", grub_cmd_sataled_ctrl,
				     GRUB_COMMAND_FLAG_BOTH,
				     "sataled_ctrl [-12345678abcdefgh] on|off|blink",
				     "Turn on/off SATA LED", sataled_ctrl_options);

	cmd_lcd_set_buff =
		grub_register_extcmd("lcd_set_buff", grub_cmd_lcd_set_buff,
				     GRUB_COMMAND_FLAG_BOTH,
				     "lcd_set_buff [-f first column] [-s second column] buffer",
				     "Set LCD buffer string.", lcd_set_buff_options);

	cmd_lcd_disp_ctrl =
		grub_register_extcmd("lcd_disp_ctrl", grub_cmd_lcd_disp_ctrl,
				     GRUB_COMMAND_FLAG_BOTH,
				     "lcd_disp_ctrl [-lchod01234567] on|off",
				     "Set ON/OFF for LCD items.", lcd_disp_ctrl_options);

	cmd_lcd_disp_item =
		grub_register_extcmd("lcd_disp_item", grub_cmd_lcd_disp_item,
				     GRUB_COMMAND_FLAG_BOTH,
				     "lcd_disp_item link|hostname|ipaddress|opmode|date|buffer0|buffer1|buffer2|buffer3|buffer4|buffer5|buffer6|buffer7",
				     "Force display item.", 0);

	cmd_lcd_disp_item =
		grub_register_extcmd("led_brightness", grub_cmd_led_brightness,
				     GRUB_COMMAND_FLAG_BOTH,
				     "led_brightness [-b 0-15] [-l 0-15]",
				     "change brightness.", led_brightness_options);

	/* Set default settings.  */
	uart_settings.port = uart_get_port(1);
	uart_settings.divisor = uart_get_divisor(38400);
	uart_settings.word_len = UART_8BITS_WORD;
	uart_settings.parity = UART_EVEN_PARITY;
	uart_settings.stop_bits = UART_1_STOP_BIT;

	uart_init();
	send_preamble();
}

GRUB_MOD_FINI(buffalo_micon)
{
	grub_unregister_extcmd(cmd);
	grub_unregister_extcmd(cmd_bz_on);
	grub_unregister_extcmd(cmd_led_ctrl);
	grub_unregister_extcmd(cmd_sataled_ctrl);
	grub_unregister_extcmd(cmd_lcd_set_buff);
	grub_unregister_extcmd(cmd_lcd_disp_ctrl);
	grub_unregister_extcmd(cmd_lcd_disp_item);
}
